/*
 state.c

 * MacFCEU 0.6.1
 *
 * Copyright notice for this file:
 *  state.c Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef mac
 #include <fcntl.h>
#else
 #include <unistd.h>
#endif

#include "driver.h"
#include "x6502.h"
#include "FCEU.h"
#include "PPU.h"
#include "FDS.h"
#include "GameGenie.h"
#include "fsound.h"
#include "general.h"
#include "state.h"
#include "movie.h"
#include "memory.h"
#include "netplay.h"
#include "video.h"
#include "utility/endian.h"

#ifdef env68k
 #pragma segment files
#endif

static void (*SPreSave)(void);
static void (*SPostSave)(void);

static int SaveStateStatus[10];
static int StateShow;

void (*FBO)(uint8*,uint32) = FlipByteOrder;

static SFORMAT SFMDATA[64];
static int SFEXINDEX;

#define RLSB 		FCEUSTATE_RLSB	//0x80000000

extern SFORMAT FCEUPPU_STATEINFO[];
extern SFORMAT FCEUSND_STATEINFO[];
extern SFORMAT FCEUCTRL_STATEINFO[];

SFORMAT SFCPU[] =
{
 { &XFrame.PC, 2|RLSB, "PC\0"},
 { &XFrame.A, 1, "A\0\0"},
 { &XFrame.P, 1, "P\0\0"},
 { &XFrame.X, 1, "X\0\0"},
 { &XFrame.Y, 1, "Y\0\0"},
 { &XFrame.S, 1, "S\0\0"},
 { RAM, 0x800, "RAM"},
 { 0 }
};

SFORMAT SFCPUC[] =
{
 { &XFrame.jammed, 1, "JAMM"},
 { &XFrame.IRQlow, 4|RLSB, "IQLB"},
 { &XFrame.tcount, 4|RLSB, "ICoa"},
 { &XFrame.count,  4|RLSB, "ICou"},
 { &timestampbase, sizeof(timestampbase)|RLSB, "TSBS"},
 { &XFrame.mooPI, 1, "PI\0\0"}, //ends "Quick and dirty hack."; 2010.03.08
 { 0 }
};

/* ---- code */

static void FlipByteOrderBroken(uint8 *s,
	uint32 c)
{ 
	uint8	*e, t;

	if (s==0 || c==0 || c&0x80000001U)
		return;
	e = c+s;
	do
		t = *--e, *e = *s, *s++ = t;
	while(--c!=0);
}

static int SubWrite(FILE *st,
	SFORMAT *sf)
{
	uint32	acc = 0;

	for(; sf->v; ++sf)
	{
		int32	z;

		if (sf->s==~0)	//Link to another struct.
		{
			uint32	tmp;

			if ((tmp=SubWrite(st,(SFORMAT*)sf->v))==0)
				return 0;
			acc+= tmp;
			continue;
		}

		acc+= 8;	/* Description + size */
		acc+= sf->s&(~RLSB);

		if (st==0)
			continue; //only calculating size of block
		//writing block
		z = sf->s&(~RLSB);

		fwrite(sf->desc,1,4,st);
		if (!write32le(z,st))
			return 0;
	//temporarily reverse byte order for big endian systems
	#ifndef LSB_FIRST
		if (sf->s&RLSB)
			FBO(sf->v,z);
	#endif
		if (fwrite((uint8*)sf->v,1,z,st)!=z)
			return 0;
	//restore byte order
	#ifndef LSB_FIRST
		if (sf->s&RLSB)
			FBO(sf->v,z);
	#endif
	}

	return acc;
}

static int WriteStateChunk(FILE *st,
	int type,
	SFORMAT *sf)
{
	int	bsize;

	fputc(type,st);

	bsize = SubWrite(0,sf);
	if (write32le(bsize,st)==0 || SubWrite(st,sf)==0)
		return 0;

	return bsize+5;
}

static SFORMAT *CheckS(SFORMAT *sf,
	uint32 tsize,
	const char *desc)
{
	for(; sf->v; ++sf)
	{
		int	z;

		if (sf->s==~0)	/* Link to another SFORMAT structure. */
		{
			SFORMAT	*tmp;

			if ((tmp=CheckS((SFORMAT*)sf->v,tsize,desc))!=0)
				return tmp;
			continue;
		}
		z = strlen(sf->desc)+1;
		if (!memcmp(desc,sf->desc,z>3 ? 4:z)) //z>7 ? 8:z
		{
			if (tsize!=(sf->s&~RLSB))
				return 0;
			return sf;
		}
	}

	return 0;
}

static int ReadStateChunk(FILE *st,
	SFORMAT *sf,
	int size)
{
	SFORMAT	*tmp;
	int	blockend = size+ftell(st);

	while(ftell(st)<blockend)
	{
		uint32	tsize, z;
		char	toa[4];

		if (fread(toa,1,4,st)<=0
				|| read32le(&tsize,st)==0)
			return 0;

		if ((tmp=CheckS(sf,tsize,toa))==0)
		{
			fseek(st,tsize,SEEK_CUR);
			continue;
		}

		z = tmp->s&(~RLSB);

		fread((uint8*)tmp->v,1,z,st);
	#ifndef LSB_FIRST
		if (tmp->s&RLSB)
			FBO(tmp->v,z);
	#endif
	}

	return 1;
}

static int ReadStateChunks(FILE *st,
	int32 totalsize)
{
	uint32	size;
	int	t, ret = 1, warn = 0;

	while(totalsize>0)
	{
		if ((t=fgetc(st))==EOF //type byte
				|| read32le(&size,st)==0)
			break;
		totalsize-= size+5;

		if (--t==0) //1
		{
			ret = ReadStateChunk(st,SFCPU,size);
			//prevent possible hang-time; 2010.04.03, 2010.04.05 (was +/-10000)
			if ((unsigned)(XFrame.count+0x800000)>0x1000000)
				XFrame.count = 0, warn = 1;
			if ((unsigned)(XFrame.tcount+0x800000)>0x1000000)
				XFrame.tcount = 0, warn = 1;
		} else if (--t==0) //2
		{
			XFrame.mooPI = XFrame.P; //in case it's not set (old state)
			ret = ReadStateChunk(st,SFCPUC,size);
		} else if (--t==0) //3
			ret = ReadStateChunk(st,FCEUPPU_STATEINFO,size);
		else if (--t==0) //4
			ret = ReadStateChunk(st,FCEUCTRL_STATEINFO,size);
		else if (--t==0) //5
			ret = ReadStateChunk(st,FCEUSND_STATEINFO,size);
		else if ((t-=0x10-5)==0)
			ret = ReadStateChunk(st,SFMDATA,size);
		else if (fseek(st,size,SEEK_CUR)<0)
			break;
	}

	if (warn)
		FCEUD_Message("##Loaded state file appears to be corrupt...");

	return ret;
}

int	CurrentState = 0;

int FCEUSS_SaveFP(FILE *st)
{
	static uint32	totalsize;
	static uint8	header[16] = "FCS\377";

	FBO = FlipByteOrder; //2010.04.03

	memset(header+4,0,12); //length [in FCEU 0.98.13] was 13!

	FCEU_en32lsb(8+header,FCEU_VERSION_NUMERIC);
	fwrite(header,1,16,st);
	FCEUPPU_SaveState();
	FCEUSND_SaveState();

	totalsize = WriteStateChunk(st,1,SFCPU);
	totalsize+= WriteStateChunk(st,2,SFCPUC);
	totalsize+= WriteStateChunk(st,3,FCEUPPU_STATEINFO);
	totalsize+= WriteStateChunk(st,4,FCEUCTRL_STATEINFO);
	totalsize+= WriteStateChunk(st,5,FCEUSND_STATEINFO);

	if (SPreSave)
		SPreSave();
	totalsize+= WriteStateChunk(st,0x10,SFMDATA);
	if (SPreSave)
		SPostSave();

	fseek(st,4,SEEK_SET);

	return write32le(totalsize,st) ? 1:0;
}

void FCEUSS_Save(char *fname)
{ /* fixed CurrentState display showing when filename is passed; 2009.10.07 03:45 EDT */
	FILE	*st = 0;
	char	*fn;

	if (GenieStage==1)
	{
		FCEU_DispMessage("Cannot save FCS in GG screen.");
		return;
	}

	if (fname)
		st = FCEUD_UTF8fopen(fname,"wb");
	else if ((fn=FCEU_MakeFName(FCEUMKF_STATE,CurrentState,0))!=0)
	{
		st = FCEUD_UTF8fopen(fn,"wb");
		free(fn);
	}
	if (st==0)
	{
		FCEU_DispMessage(fname ? "State save error.":"State %d save error.",
			CurrentState);
		return;
	}

	FCEUSS_SaveFP(st);

	SaveStateStatus[CurrentState] = 1;
	fclose(st);
	FCEU_DispMessage(fname ? "State saved.":"State %d saved.",
		CurrentState);
}

int FCEUSS_LoadFP(FILE *st)
{
	uint8	header[16];
	int	x, stateversion;

	if (fread(&header,1,16,st)!=16
			|| header[0]!='F' || header[1]!='C' || header[2]!='S')
		return 0;

	if (header[3]==0xFF)
		stateversion = FCEU_de32lsb(8+header);
	else
		stateversion = header[3]*100;
	//MacFCEU 0.6a3 and earlier bug fix (oh, boy...)
#ifdef mac
	if ((stateversion-9813)<=1)
		FBO = FlipByteOrderBroken;
	else
		FBO = FlipByteOrder;
#endif

	//FCE Ultra bug fix - stored LE, read without endian conversion; 2008.09.11
	x = ReadStateChunks(st,FCEU_de32lsb(header+4));
	if (stateversion<9500)
		XFrame.IRQlow = 0;
	if (GameStateRestore)
		GameStateRestore(stateversion);
	if (x!=0)
	{
		FCEUPPU_LoadState(stateversion);
		FCEUSND_LoadState(stateversion);
	}

	return x;
}

int FCEUSS_Load(char *fname)
{ /* CurrentState display...; 2009.10.07 07:27 EDT */
	FILE	*st;
	char	*fn;

	if (GenieStage==1)
	{
		FCEU_DispMessage("Cannot load FCS in GG screen.");
		goto endo;
	}

	if (fname)
		st = FCEUD_UTF8fopen(fname,"rb");
	else if ((fn=FCEU_MakeFName(FCEUMKF_STATE,CurrentState,fname))!=0)
	{
		st = FCEUD_UTF8fopen(fn,"rb");
		free(fn);
	} else
		st = 0;

	if (st==0)
	{
		if (fname)
			FCEU_DispMessage("State load error.",
				CurrentState);
		else
		{
			FCEU_DispMessage("State %d load error.",
				CurrentState);
			SaveStateStatus[CurrentState] = 0;
		}
		goto endo;
	}

	if (FCEUSS_LoadFP(st))
	{
		if (fname)
			FCEU_DispMessage("State loaded.");
		else
		{
			FCEU_DispMessage("State %d loaded.",
				CurrentState);
			SaveStateStatus[CurrentState] = 1;
		}
		fclose(st);
		return 1;
	} else
	{
		SaveStateStatus[CurrentState] = 1; //does this make any sense?
		FCEU_DispMessage(fname ?
			"Error(s) reading state file":"Error(s) reading state %d!",
			CurrentState);
		fclose(st);
	}
endo:
	return 0;
}

void FCEUSS_CheckStates(void)
{
	char	*fn;
	int	ssel;

	for(ssel=0; ssel<10; ++ssel)
	{
		if ((fn=FCEU_MakeFName(FCEUMKF_STATE,ssel,0))!=0
				&& CheckFileExists(fn))
			SaveStateStatus[ssel] = 1;
		else
			SaveStateStatus[ssel] = 0;
		if (fn)
			free(fn);
	}
	CurrentState = 0;
	StateShow = 0;
}

void ResetExState(void (*PreSave)(void),
	void (*PostSave)(void))
{
	int	x;

	for(x=0; x<SFEXINDEX; ++x)
		SFMDATA[x].desc[0] = 0;
	SFMDATA[0].s = 0; //initialize

	SPreSave = PreSave;
	SPostSave = PostSave;
	SFEXINDEX = 0;
}

void AddExState(void *v,
	uint32 s,
	int type,
	const char *desc)
{
	if (desc)
		memcpy(SFMDATA[SFEXINDEX].desc,desc,8);
	else
		SFMDATA[SFEXINDEX].desc[0] = 0;
	SFMDATA[SFEXINDEX].v = v;
	SFMDATA[SFEXINDEX].s = s;
	if (type)
		SFMDATA[SFEXINDEX].s|= RLSB;
	if (SFEXINDEX<63)
		++SFEXINDEX;
	SFMDATA[SFEXINDEX].v = 0;	// End marker.
}

void FCEUI_SelectState(int w)
{
	if (w==-1)
		{StateShow = 0; return;}
	FCEUI_SelectMovie(-1);

	CurrentState = w;
	StateShow = 180;
	FCEU_DispMessage("-select state-");
}

void FCEUI_SaveState(char *fname)
{
	StateShow = 0;
	FCEUSS_Save(fname);
}

void FCEUI_LoadState(char *fname)
{
	StateShow = 0;

	FCEUMOV_Stop();

 /* For network play, be the state loaded locally, and then save the state to a temporary file,
	and send that.  This insures that if an older state is loaded that is missing some
	information expected in newer save states, desynchronization won't occur
	(at least not from this ;) ).
 */
	if (FCEUSS_Load(fname) && FCEUnetplay)
	{
		char	*fn = FCEU_MakeFName(FCEUMKF_NPTEMP,0,0);
		FILE	*fp;

		if ((fp=fopen(fn,"wb"))!=0)
		{
			int	r = FCEUSS_SaveFP(fp);

			fclose(fp);
			if (r)
				FCEUNET_SendFile(FCEUNPCMD_LOADSTATE,fn);
			unlink(fn);
		}
		free(fn);
	}
}

void FCEU_DrawSaveStates(uint8 *XBuf)
{
	if (!StateShow)
		return;
	--StateShow;
	FCEU_DrawNumberRow(XBuf,SaveStateStatus,CurrentState);
}
